function A = herml(C, X, A_0, ns, params, W)
% A = herml(C, X, A_0, params)
%
% Core herml learning algorithm
% 
% C: 4 column matrix
%      column 1, 2: index of constrained points.  Indexes between 1 and n
%      column 3: 1 if points are similar, -1 if dissimilar
%      column 4: right-hand side (lower or upper bound, depending on 
%                   whether points are similar or dissimilar)
%
% X: (n x m) data matrix - each row corresponds to a single instance
%
% A_0: (m x m) regularization matrix
%
% params: algorithm parameters - see see SetDefaultParams for defaults
%           params.thresh: algorithm convergence threshold
%           params.gamma: gamma value for slack variables
%           params.max_iters: maximum number of iterations
%
% returns A: learned Mahalanobis matrix

tol = params.thresh;
gamma = params.gamma;
max_iters = params.max_iters;


i = 1;
iter = 0;
c = size(C, 1);
lambda = zeros(c,1);
bhat = C(:,4);
lambdaold = zeros(c,1);
conv = Inf;
A = A_0;
p = zeros(ns,1);
v = cell(ns,1);

while (true),
    i1 = C(i,1);
    i2 = C(i,2);
    
    for ii = 1 : ns
        X_s = X{ii};
        v{ii} = X_s(i1,:)' - X_s(i2,:)';
        p(ii) = v{ii}'*A{ii}*v{ii};
    end

    if (abs(bhat(i)) < 10e-10),
        error('bhat should never be 0!');
    end
  
    
    delta = C(i,3);
    alpha = solve_alpha(p',bhat(i),gamma,delta,ns,W(1:ns));

    alpha = min(lambda(i),alpha);  
    lambda(i) = lambda(i) - alpha;
    bhat(i) = gamma* bhat(i)/(gamma + delta*alpha*bhat(i)); 
    
    for ii = 1 : ns
        beta = delta*alpha/(1 - delta*alpha*p(ii));
        A{ii} = A{ii} + (beta*A{ii}*v{ii}*v{ii}'*A{ii});
    end 

    if i == c
        normsum = norm(lambda) + norm(lambdaold);
        if (normsum == 0)
            break;
        else
            conv = norm(lambdaold - lambda, 1) / normsum;
            if (conv < tol || iter > max_iters),
                break;
            end
        end
        lambdaold = lambda;
    end
    i = mod(i,c) + 1;

    iter = iter + 1;
    if (mod(iter, 5000) == 0),       
       disp(sprintf('herml iter: %d, conv = %f', iter, conv));
    end
end
disp(sprintf('herml converged to tol: %f, iter: %d', conv, iter));

end

%% //////////////////////////////////////////////////////////////
% solve alpha
function [alpha]=solve_alpha(P,epsi,gamma,delta,nblocks,W)

options = optimset('Display', 'off','MaxIter',200); % Turn off Display
alpha= fsolve(@myfun, 0, options);
function F=myfun(x)
F=((P./(ones(1,nblocks)-(delta*P*x)))*W)/nblocks-(gamma* epsi/(gamma + delta*x*epsi));%(epsi-(delta/gamma)*x);
end
end
